/*
 * display_data.cpp
 *
 *  Created on: 18 mars 2015
 *      Author: dominique
 */


#include <cstdlib>
#include <fstream>

#include <algorithm>
#include "database.hpp"
#include "utils.hpp"
#include "display_data.hpp"



using namespace std ;




/**
 *   affiche chaque transaction désignés par list_objects
 *   (les noms des objets initiaux de chaque transaction sont affichés)
 */
void display_objects_names(
							const set<int>& initial_objects,
							const vector<string>& initial_objects_names)
{
	// displayVector("initial_objects", initial_objects) ;
	// displayVector("objects_names", initial_objects_names) ;

	cout << "{" ;

	if (initial_objects.size() > 0)
	{
		set<int>::const_iterator it = initial_objects.begin() ;

		cout << "\"" << initial_objects_names.at(*it) << "\"";
		++it ;

		for ( ; it != initial_objects.end() ; ++it)
		{
			cout << ", \"" << initial_objects_names.at(*it) << "\"";
		}
	}

	cout << "}" ;
}



/**
 *   affiche chaque transaction désignés par list_objects
 *   (les noms des objets initiaux de chaque transaction sont affichés)
 */
void display_objects_names(
							const vector<int>& initial_objects,
							const vector<string>& initial_objects_names)
{
	// displayVector("initial_objects", initial_objects) ;
	// displayVector("objects_names", initial_objects_names) ;

	cout << "{" ;

	if (!initial_objects.empty())
	{
		vector<int>::const_iterator it = initial_objects.begin() ;

		cout << "\"" << initial_objects_names.at(*it) << "\"";
		++it ;

		for ( ; it != initial_objects.end() ; ++it)
		{
			cout << ", \"" << initial_objects_names.at(*it) << "\"";
		}
	}

	cout << "}" ;
}



/**
 *   affiche chaque transaction désignée par list_objects
 *   (les noms des objets initiaux de chaque transaction sont affichés)
 */
void display_n_objects_names(
		                    int n,
		 	 	 	 	 	const TransactionTable& tt,
							const vector<int>& list_transactions,
							const vector<string>& initial_objects_names)
{
	unsigned int i = 0 ;

	while(i < list_transactions.size() && i < n)
	{

        cout << list_transactions.at(i) ;
		display_objects_names(
								tt.at(list_transactions.at(i)).initial_objects,
								initial_objects_names
							  ) ;
		cout << " " ;

		i++ ;
	}
	cout << endl ;
}




/**
 *   affiche le premier triangle de la liste
 */
void display_first_triangle(
		 	 	 	 	 	const TransactionTable& tt,
							const vector<int>& list_transactions,
							const vector<string>& initial_objects_names)
{
	// displayVector("list_transactions", list_transactions) ;
	// displayVector("objects_names", initial_objects_names) ;


        // cout << list_transactions.at(0) ;
		display_objects_names(
								tt.at(list_transactions.at(0)).initial_objects,
								initial_objects_names
							  ) ;
}





/**
 *   affiche chaque transaction désignée par list_objects
 *   (les noms des objets initiaux de chaque transaction sont affichés)
 */
void display_objects_names(
		 	 	 	 	 	const TransactionTable& tt,
							const vector<int>& list_transactions,
							const vector<string>& initial_objects_names)
{
	// displayVector("list_transactions", list_transactions) ;
	// displayVector("objects_names", initial_objects_names) ;

	for (unsigned int i = 0 ; i < list_transactions.size() ; i++)
	{

        cout << list_transactions.at(i) ;
		display_objects_names(
								tt.at(list_transactions.at(i)).initial_objects,
								initial_objects_names
							  ) ;
		cout << " " ;
	}
	cout << endl ;
}


void outputOriginalPattern(
							bool show_tids,
							TransactionTable& tt,
							const Transaction& occs,
							set_t pattern,
							int set_support,
							const vector<string>& items_names
				  )
{
					  // output lclosed pattern
		#ifdef TRACK_TIDS
						// Get original tids
		set_t orig_tids;

		if (show_tids)
		{
			orig_tids.reserve(occs.size());
			for(unsigned int i = 0; i < occs.size(); i++)
			{
				orig_tids.insert(orig_tids.end(), tt[occs[i]].tids.begin(),
											  tt[occs[i]].tids.end());
			}
			sort(orig_tids.begin(), orig_tids.end());
		}

		original_pattern_print(pattern, set_support, orig_tids, items_names);

		#else
		set_t *orig_tids; //dummy set, won't be read
		pattern_print(closed_pattern, set_support, *orig_tids);

		#endif //TRACK_TIDS
}





void outputPattern(
				   bool show_tids,
				   TransactionTable& tt, const Transaction& occs,
				   set_t pattern, int set_support
				  )
{
					  // output lclosed pattern
		#ifdef TRACK_TIDS
						// Get original tids
		set_t orig_tids;
		if (show_tids)
		{
			orig_tids.reserve(occs.size());
			for(unsigned int i = 0; i < occs.size(); i++)
			{
				orig_tids.insert(orig_tids.end(), tt[occs[i]].tids.begin(),
											  tt[occs[i]].tids.end());
			}
			sort(orig_tids.begin(), orig_tids.end());
		}

		pattern_print(pattern, set_support, orig_tids);

		#else
		set_t *orig_tids; //dummy set, won't be read
		pattern_print(closed_pattern, set_support, *orig_tids);

		#endif //TRACK_TIDS
}


void outputLocalPattern(
				   	   	 bool show_tids,
						 const TransactionTable& tt,
						 const vector<int>& occs_cci,
						 set_t closed_pattern_cci,
						 int set_support_cci,
						 const vector<string>& items_names
						)
{
			  	  // output local closed pattern
	#ifdef TRACK_TIDS
		// Get original tids
		set_t orig_tids;
		if (show_tids)
		{
			orig_tids.reserve(occs_cci.size());

			for (vector<int>::const_iterator it = occs_cci.begin() ;
				     it != occs_cci.end() ; ++it)
			{
				orig_tids.insert(orig_tids.end(), tt[*it].tids.begin(),
						         tt[*it].tids.end());
			}
			std::sort(orig_tids.begin(), orig_tids.end());
		}
		original_pattern_print(closed_pattern_cci, set_support_cci,
				               orig_tids, items_names);

	#else
		set_t *orig_tids; //dummy set, won't be read
		original_pattern_print(closed_pattern, membership_retval,
				               *orig_tids, items_names);
	#endif
}

	/**
	 * each Transaction of tt contains a list of adjacences.
	 * each list of adjacences contains the indices of the transactions in tt
	 * that are connected to the current transaction
	 */
void print_adjacences(TransactionTable tt)
{
	cout << endl ;

	for (unsigned int i = 0 ; i < tt.size() ; i++)
	{
		cout << "adjacences transaction " << i << " : " ;

		for (vector<int>::const_iterator j = tt.at(i).adjacences.begin() ;
			  j != tt.at(i).adjacences.end() ; ++j)
		{
			cout << *j << " " ;
		}
		cout << endl ;
	}
}


/**
 * print the initial objects adajacent to the object described by
 * the transaction t
 */
void print_initial_objects_transaction(const Transaction& t)
{
	for (vector<int>::const_iterator it = t.initial_objects.begin() ;
	     it != t.initial_objects.end() ; ++it)
	{
		cout << *it << " " ;
	}
	cout << endl ;
}


/**
 * each Transaction of tt contains a list of adjacences.
 * each list of adjacences contains the indices of the transactions in tt
 * that are connected to the current transaction
 */
void print_initial_objects(const TransactionTable& tt)
{
	cout << endl ;

	for (unsigned int i = 0 ; i < tt.size() ; i++)
	{
		cout << "objects transaction  " << i << " : " ;
		print_initial_objects_transaction(tt.at(i)) ;
	}
}


/**
 * each Transaction of tt contains a list of adjacences.
 * each list of adjacences contains the indices of the transactions in tt
 * that are connected to the current transaction
 */
void print_selection_initial_objects(const TransactionTable& tt,
									 const vector<element_t>& list)
{
	for (vector<element_t>::const_iterator i = list.begin() ;
		  i != list.end() ; ++i)
	{
		print_initial_objects_transaction(tt.at(*i)) ;
	}
}


