/*
 * graphe.hpp
 *
 *  Created on: 11 avr. 2015
 *      Author: Dominique Bouthinon
 */

#ifndef SRC_GRAPH_HPP_
#define SRC_GRAPH_HPP_

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/connected_components.hpp>

using namespace std;
using namespace boost ;


struct VertexProperties
{
   unsigned ident;
   VertexProperties() : ident(0) {}
   VertexProperties(unsigned i) : ident(i) {}
};


struct EdgeProperties
{
   float weight;
   float distance;
   EdgeProperties() : weight(0.0), distance(0.0) {}
   EdgeProperties(float w, float d) : weight(w), distance(d) {}
};


struct GraphProperties
{
   string name ;
   GraphProperties() : name("") {}
   GraphProperties(string n) : name(n) {}
};


typedef property<edge_weight_t, int> EdgeWeightProperty;

typedef boost::adjacency_list<
							   vecS, vecS, undirectedS,
							   VertexProperties
							   //EdgeProperties>,
							   //GraphProperties>
							 > Graphe;


typedef graph_traits<Graphe>::vertex_descriptor vertex_descriptor ;
typedef graph_traits<Graphe>::edge_descriptor edge_descriptor ;

typedef graph_traits<Graphe>::edge_iterator edge_iterator ;
typedef graph_traits<Graphe>::vertex_iterator vertex_iterator ;
typedef graph_traits<Graphe>::adjacency_iterator adjacency_iterator ;
typedef graph_traits<Graphe>::out_edge_iterator out_edge_iterator ;
typedef graph_traits<Graphe>::in_edge_iterator in_edge_iterator ;

/*
typedef filtered_graph
		<
		 Graphe,
		 function<bool(edge_descriptor)>,
		 function<bool(vertex_descriptor)>
        >
	    ComponentGraph;
	    */


/* ============================== FUNCTIONS ================================= */

void displayVertices(Graphe& g) ;
void displayEdges(Graphe& g) ;


#endif /* SRC_GRAPH_HPP_ */



