/**
 * class template graph.hpp
 *
 * Dominique Bouthinon
 * LIPN, Université Paris 13 (France)
 * db@lipn.univ-paris13.fr
 *
 * represents a matrix
 * a row i only contains the nodes connected with the node i
 * (so rows have different sizes)
 *
 * This class can be used to represent non rectangular matrices
 * and rectangulare ones
 *
 * (the explicit use of this-> is  volunteer for sake of readability)
 */

#ifndef   	_MATRIX_HPP_
#define   	_MATRIX_HPP_


#include <iostream>
#include <vector>
#include <sstream>
#include <fstream>

#include "row_matrix.hpp"

using namespace std ;


template <typename T> class matrix: public vector<row_matrix<T> >
{

	public:

		// --------------------------- MATRIX HANDING --------------------------

						// constructeur
		matrix() ;

						// return the number of rows
		int numberRows() ;

						// display the matrix
		void displayMatrix() ;

						// read values of the matrix in the file FileMatrix
		void readValuesMatrix(const char *fileMatrix) ;


		// ------------------------ ROW HANDING --------------------------------

						// return the number of elements row i
		int numberElementsRow(int i) ;

						// display the row i
		void displayRow(int i) ;

						// add an empty row at the end of the matrix
		void addRow() ;

						// add an empty row at position i of the matrix
		void addRow(int i) ;

						// add row at the end of the matrix
		void addRow(row_matrix<T> row) ;

						// return the row number i
		row_matrix<T> getRow(int i) ;

						// delete the row i (which no longer exists)
		void deleteRow(int i) ;

						// clear the row i (which becomes empty)
		void clearRow(int i) ;


		// --------------------- ELEMENT HANDING -------------------------------

						// add e at the end of the row i
		void addElement(T e, int i) ;

						// add e at the end of the row i, column j
		void addElement(T e, int i, int j) ;

						// return the element of coordinates (i,j)
		T getElement(int i, int j) ;

						// delete the element (i,j)
		void deleteElement(int i, int j) ;
};


// ====================== DEFINITION METHODS ==================================


// --------------------------- MATRIX HANDLING --------------------------------

					// constructor
template <typename T> matrix<T>::matrix():vector<row_matrix<T> >()
{
}

					// return the number of arrows
template <typename T> int matrix<T>::numberRows()
{
	return this->size() ;
}

					// display the matrix
template <typename T> void matrix<T>::displayMatrix()
{
	for (int i = 0 ; i < this->numberRows() ; i++)
	{
		cout << "row " << i <<  " : " ;
		this->displayRow(i) ;
	}
}


					// read values of the matrix in the file FileMatrix
template <typename T> void matrix<T>::readValuesMatrix(const char *fileMatrix)
{

	ifstream ifs (fileMatrix , ifstream::in);

	int i = 0 ;						// number of the current line
	while (ifs.good())
	{
		string line;
		stringstream ss;

		T value ;

		ss<<skipws;

		getline(ifs, line);

		// cout << "line =" << line << "x" << endl ;

		if (line != "")
		{
			ss<<line;

			this->addRow() ;				// new row for the matrix

			do
			{
				ss >> value ;				// current value read

				if(!ss.fail())
				this->addElement(value, i) ;	// add value to row i
			}
			while(ss.good());
		}
		i++ ;							// next line
	}
}




// ---------------------------- ROW HANDLING ----------------------------------

					// add an empty row at the end of the matrix
template <typename T> void matrix<T>::addRow()
{
	this->push_back(row_matrix<T>()) ;
}


					// add an empty row at position i of the matrix
template <typename T> void matrix<T>::addRow(int i)
{
	this->insert(row_matrix<T>(), this->begin() + i) ;
}


						// add a row at the end of the matrix
template <typename T> void matrix<T>::addRow(row_matrix<T> row)
{
	this->push_back(row) ;
}


						// return the row number i
template <typename T> row_matrix<T> matrix<T>::getRow(int i)
{
	return this->at(i) ;
}

						// delete the row i (which no longer exists)
template <typename T> void matrix<T>::deleteRow(int i)
{
	this->erase(this->begin() + i) ;
}


						// clear the row i (that becomes empty)
template <typename T> void matrix<T>::clearRow(int i)
{
	this->at(i).clearRow() ;
}


						// return the number of elements in row i
template <typename T> int matrix<T>::numberElementsRow(int i)
{
	return this->at(i).numberElements() ;
}

						// display the row i
template <typename T> void matrix<T>::displayRow(int i)
{
	this->getRow(i).displayRow() ;
}




//---------------------------- ELEMENT HANDLING --------------------------------

						// add e at the end of the row i
template <typename T> void matrix<T>::addElement(T e, int i)
{
	this->at(i).addElement(e) ;
}


						// add e at row i, column j
template <typename T> void matrix<T>::addElement(T e, int i, int j)
{
	this->at(i).addElement(e, j) ;
}


						// return the element of coordinates (i,j)
template <typename T> T matrix<T>::getElement(int i, int j)
{
	return this->at(i).getElement(j) ;
}

						// delete the element of coordinates (i,j)
template <typename T> void matrix<T>::deleteElement(int i, int j)
{
	this->at(i).deleteElement(j) ;
}


#endif

//==================== END CLASS TEMPLATE matrix.hpp =======================


