/*
** paraminer.cpp
** 
** Made by (Benjamin Negrevergne)
** Email   [firstname].[name]@imag.fr
** 
** Started on  Thu Sep  2 10:21:29 2010 Benjamin Negrevergne
**
** * PARAMINER LOCAL CLOSURES
 * (by Henry SOLDANO (LIPN University Paris 13)
 * February 2015
 *
 * program modified by Dominique Bouthinon (LIPN, University Paris 13)
*/


#include <iostream>
// #include <algorithm>
#include <map>
#include <cstdio>
#include <cctype>
#include <cmath>
#include <pthread.h>
#include <unistd.h>
#include <limits>
#include <map>

#include "paraminer_LC2.hpp"
#include "paraminer_local.hpp"
#include "utils.hpp"
#include "element.hpp"
#include "pattern.hpp"
#include "database.hpp"
#include "display_data.hpp"


using namespace boost ;


int num_threads = 1; 

#define PARALLEL_PROCESS

#ifdef PARALLEL_PROCESS

	extern "C"
	{
		#include "tuplespace.h"
		#include "thread.h"
	}

	#include "melinda_local.hpp"

	tuplespace_t ts;

	int depth_tuple_cutoff = 2;

	float max_reduction_threshold = 10; 	/**< \brief Threshold to
	 control activation of el-reduction on level 0 calls to
	 expand. (-y switch.)

	 On level 0 expand calls, el-reduction is activated by
	 default. The reductions are performed unless the maximum
	 reduction factor is garanteed to be below
	 max_reduction_factor_threshold. */

	float min_reduction_threshold = 2; /**< \brief Threshold to control
	 activationi of el-reduction on level 1 and above. (-x switch.)
	 
	 On level 1 and above, el-reduction is deactivated by
	 default. The el-reductions are activated if the minimum
	 reduction factor is garanteed to be above
	 min_reduction_factor_threshold.
 */

#endif //PARALLEL_PROCESS

using std::cout; 
using std::endl;
using std::cerr; 
using std::pair;

using namespace boost ;

TransactionTable &tt = *new TransactionTable;
TransactionTable &ot = *new TransactionTable; 

int threshold; 

bool show_tids = false ;


struct support_sort_cmp_t
{
  const TransactionTable &ot;
  bool operator()(int a, int b)
  {
    return ot[a].size() < ot[b].size(); 
  }
};



/**
 * return the identifiers of the transactions in occs
 */
void ident_transactions2(const Transaction& occs, map<int,bool>& id)
{
	for (unsigned int i = 0 ; i < occs.size() ; ++i)
	{
		id[occs.at(i)] = true ;
	}
}



/**
 * Calcul le graphe induit par l'extension du pattern clos "pattern"
 * Pour des raisons de sécurité dues au parallélisme de paraminer,
 * on fait une première version en générant un graphe, sans utiliser de mises
 * à jour d'un graph précédent.
 */
void compute_induced_graph2( int depth,			      // depth recursion
							 Graphe& g,
							 TransactionTable &tt,    // table transactions
							 const Transaction occs  // numéros des transactions
													 // contenant le pattern clos courant
					      )
{

	map<int,bool> id_trans ;

	ident_transactions2(occs, id_trans) ;
											// les identifiants sont classés pour
											// eviter de creer des noeuds doublons
											// dans le graphe induit

											// contiendra les noeuds déjà créés.
	map<int,vertex_descriptor> created_nodes ;


							// pour chaque identifiant de transaction dans id_trans
	for (map<int, bool>::iterator ite_trans = id_trans.begin() ; ite_trans != id_trans.end() ; ite_trans++)
	{
		int id = (*ite_trans).first ;	   // id is a transaction number

		if (created_nodes.find(id) == created_nodes.end())
		{
							// le noeud id n'existe pas on le créé
			created_nodes[id] = add_vertex(VertexProperties(id), g) ;
		}

		vertex_descriptor vid = created_nodes[id] ;

		// VertexProperties const& vertexProperties = g[vid];
		// cerr << "Vertex ident : " <<  vertexProperties.ident << endl;


							// pour chaque successeur du noeud courant (id)
		for (unsigned int i = 0 ; i < tt.at(id).adjacences.size() ; i++)
		{
			int id_node =  tt.at(id).adjacences.at(i) ;

			if (depth == 0 || id_trans.find(id_node) != id_trans.end()) // node must be considered
			{
				if (created_nodes.find(id_node) == created_nodes.end()) // node not created
				    created_nodes[id_node] = add_vertex(VertexProperties(id_node), g) ;

				vertex_descriptor v = created_nodes[id_node] ;

				// VertexProperties const& vertexProperties2 = g[v];
				// cerr << "Vertex ident : " <<  vertexProperties2.ident << endl;

							// creation de l'arc
				add_edge(vid, v, g) ;
			}

			// VertexProperties const& vertexProperties1 = g[v];
			// cerr << "Vertex id : " <<  vertexProperties1.id << endl;
		}
		// end for each adjacent node
	}
	// end for each id

}
// end compute_induced_graph




/** 
 * Returns true if parent pattern is the first parent of \pattern
 * according to exclusion list.  Note that given the exclusion list,
 * parent_pattern is not needed to compute this result.
 * 
 */
bool parent_pattern_is_first_parent(const set_t &pattern, const set_t &exclusion_list)
{
  assert(is_sorted(pattern));
  assert(is_sorted(exclusion_list)); 
  /* Check if one element from closed set belong to the exclusion list */ 
  set_t::const_iterator xlit = exclusion_list.begin();
  const set_t::const_iterator xlend = exclusion_list.end();
  set_t::const_iterator cit = pattern.begin();  
  const set_t::const_iterator cend = pattern.end();  
  while(xlit != xlend && cit != cend)
  {
    if(*xlit < *cit)
      ++xlit;
    else if (*xlit > *cit)
      ++cit; 
    else
    {
      return false; 
    }
  }
  return true; 
}



/**
 *  Retourne les objets initiaux (de O) impliques dans une composante
 *  connexe
 */
set<int>* involved_objects(TransactionTable &tt, const vector<int>& cci)
{
	set<int>* objets = new set<int>() ;

							// recupere les sommets
							// associés aux transactions contenues
							// dans la composante connexe de triangles cci
	for (vector<int>::const_iterator it = cci.begin() ; it != cci.end() ; ++it)
	{
								// cci.at(i) identifie une transaction
		objets->insert(tt[*it].initial_objects.begin(),
					tt[*it].initial_objects.end()) ;
	}

	return objets ;
}



/**
 * retourne true si la composante connexe cci implique au moins
 * "threshold_connecte_component" objets de O
 */
bool am(TransactionTable &tt,
		const set<int>& cci,
		int threshold_connected_component)
{

	set<int> objets ;

							// recupere les objets initiaux
							// associés aux transactions contenues
							// dans cci
	for (set<int>::const_iterator it  = cci.begin(); it != cci.end() ; ++it)
	{
								// cci.at(i) identifie une transaction
		objets.insert(tt[*it].initial_objects.begin(),
					  tt[*it].initial_objects.end()) ;
	}

	int support = objets.size() ;

	return support >= threshold_connected_component ;
}


	// compute the local closures associated with closed_pattern
int local_closures(
		            	int depth,
					TransactionTable &tt,		// transactions
					const Transaction &occs,		// id transactions contenant closed_pattern
					SupportTable support,
					element_t max_element,
					set_t pattern,
					set_t closed_pattern,
					int set_support,
					int threshold_connected_component,
					const vector<string>& initial_objects_names,
					const vector<string>& items_names
				  )
{
	int nb_local_closed_patterns1 = 0 ;

 	  	  	  	  	  	  // the current closed_pattern is ok

 	  	  	  	  	  	  /****************************************************/
 	  	  	  	  	  	  /*                    LOCAL CLOSURES                */
 	  	  	  	  	  	  /************************************************** */

 	  	  	  	  	  	  // calcul du graphe induit par ext(close_pattern)
 	  	  	  	  	  	  // occs contient les indices des transactions
 	  	  	  	  	  	  // contenant closed_pattern

 set<int>* objects_c = involved_objects(tt, occs) ;		// objets initiaux du sous-graph induit 

 Graphe* g = new Graphe;

 compute_induced_graph2(depth, *g, tt, occs) ;

 vector<int> the_components(num_vertices(*g));
 unsigned int num_components = connected_components(*g, &the_components[0]);

 	  	  	  	  	  	  // distribue les elements dans leur composante connexe
vector<int> components[num_components] ;

	// pour chaque composante connexe de g
for (unsigned int i = 0 ; i < the_components.size() ; i++)
{
	components[the_components.at(i)].push_back(i) ;
}


						// pour chaque composante connexe (triangles) de g
 for (unsigned int id_component = 0 ;  id_component < num_components ;
		id_component++)   
 {
	  vector<int> occs_cci ;						// triangles de la composante

	  	  	  	  	   // on recupere les ids des noeuds de la composante id_component
	  for (unsigned int i = 0 ; i < components[id_component].size() ; i++)
	  {
		  	  vertex_descriptor id_node = components[id_component].at(i) ;

			  VertexProperties const& vertexProperties = (*g)[id_node];
			  occs_cci.push_back(vertexProperties.ident) ;
	  }

					// on calcule si la composante connexe component
					// satisfait la condition am
					// ici la contrainte porte sur le nombre de sommes 
					// impliqués impliques dans la composante des triangles 

	  set<int>* objects_cci = involved_objects(tt, occs_cci) ;

	  if (objects_cci->size() >= threshold_connected_component)
	  {
						  // la composante connexe contient suffisamment
		  	  	  	  	  // d'objets initiaux (dans O)

						  // cci contient moins d'objets (de T) que occs
						  // on recalcule le pattern_clos à partir de
						  // ces objets

		   nb_local_closed_patterns1++ ;

		   SupportTable support_cci(max_element+1, 0);

		   	   	   	   	   // calcule les supports des items qui
		   	  	  	  	   // apparaissent dans les transactions
		   	  	  	  	   // contenant (pattern + ?) augmentation
		   	   	   	   	   // on ne tient pas compte des items dont le support
		   	   	   	   	   // est nul
		   	   	   	   	   // (un tel item soit appartient a pattern + augmentation
		   	   	   	   	   //  soit n'apparait pas dans les transactions)

		   compute_element_support_locale(
				   	   	   	   	   	   	  &support_cci,
				   	   	   	   	            support,
										  tt,
										  occs_cci
										  );

		   	   	   	   	   // pattern + augmentation présent dans toutes les
		   	   	   	   	   // transactions dénotées par occs_cci.size() ;

		  int set_support_cci = occs_cci.size() ;

		  closure_data_t c_data_cci = {tt, occs_cci, support_cci, set_support_cci};
		  set_t closed_pattern_cci  = clo(pattern, c_data_cci);

		  // cout << endl << "------------------------------------------------------" << endl ;
		  // cout << "closed pattern           : " ;

		  
		  outputOriginalPattern(false, tt, occs,closed_pattern, set_support, items_names) ;
		  cout << ", " ;

			// affiche les objets initiaux du graph
		  // display_objects_names(*objects_c, initial_objects_names) ;
		  // cout << ", " ;

			 // affiche la communaute

		  display_objects_names(*objects_cci, initial_objects_names) ;
		  cout << ", " ;

			// affiche le nombre de triangles dans la composante connexe
		  cout << occs_cci.size() << ", " ;

		  // cout << set_support << ", " ;

		  // outputOriginalPattern(show_tids, tt, occs,closed_pattern, set_support, items_names) ;
		  // cout << "connected component      : "  ;
		  // displayVectorAsSet(occs_cci) ;
		  // cout << endl ;

		  // cout << "triangles                : " ;
		  // display_objects_names(tt, occs_cci, initial_objects_names) ;

		  // cout << "number initial objects   : " << objects->size() ;
		  // cout << objects->size() << ", ";
		  // cout << endl ;
		  // cout << "local closed pattern     : " ;

 	  	  	  // affiche les sommets des triangles
		  // cout << "involved initial objects : " ;
		  // display_objects_names(*objects, initial_objects_names) ;

		
		  // affiche le nombre de triangles 
	


		  // affiche les sommets du premier triangle          " ;
		   display_first_triangle(tt, occs_cci, initial_objects_names) ;
		   cout << ", " ;

		   outputLocalPattern(false, tt, occs_cci, closed_pattern_cci, set_support_cci, items_names) ;

		  // outputLocalPattern(show_tids, tt, occs_cci, closed_pattern_cci, set_support_cci, items_names) ;
		   cout << endl ;

		  delete objects_cci ;
	  }
	  else
	  {
		  // cout << "too small community " << endl ;
		  // cout << "nb objects = " << objects->size() << endl ;
		  // cout << "involved initial objects : " ;
		  // display_objects_names(*objects, initial_objects_names) ;
		  // cout << endl ;
	  }

 }
 // end for each connected component

 delete g ;

 return nb_local_closed_patterns1 ;

 }
 // end local_closures



						// Algo 6 papier PARAMINER
size_t expand(
				TransactionTable &tt,
				TransactionTable &ot,
				const Transaction &occs,	// idents transactions contenant
											// pattern + augmentation
											// pour depth = 0 contient les
											// identifiants de toutes les transactions
				const set_t &parent_pattern,
				element_t pattern_augmentation, // vaut -1 pour depth = 0
				int depth,
				const set_t &parent_el,         // vide pour depth = 0
				const set_t &el_tail,           // vide pour depth = 0
				unsigned int threshold_connected_component,
				const vector<string>& initial_objects_names,
				const vector<string>& items_names,
				int & nb_closed_patterns,
				int & nb_rejected_closed_patterns
			 )
{

  set_t exclusion_list(parent_el.size()+el_tail.size());
  copy(parent_el.begin(), parent_el.end(), exclusion_list.begin());
  copy(el_tail.begin(), el_tail.end(), exclusion_list.begin()+parent_el.size());
  assert(is_sorted(parent_el));
  sort(exclusion_list.begin(), exclusion_list.end()); //TODO Very inefficient

  	  	  	  	  	  // exclusion_list = parent_el union el_tail

  set_t pattern(parent_pattern);

   // cout << endl ;
   // cout << "=========================================================" << endl ;

   // simple_pattern_print("parent pattern        ", pattern, items_names) ;

   /*
   if(depth > 0)
   {
	   cout << "augmentation           : " ;
	   original_element_print(pattern_augmentation, items_names) ;
	   cout << endl ;
   }
   */

  // print_transaction_table(tt) ;
  // print_transaction_table_attributes(tt,permutations) ;

  element_t max_element = tt.max_element;


  SupportTable support(max_element+1, 0);

  	  	  	  	  	  	  	  	  	  // calcule les supports des items qui
  	  	  	  	  	  	  	  	  	  // apparaissent dans les transactions
  	  	  	  	  	  	  	  	  	  // contenant (pattern + augmentation)
                                      // (ou simplement augmentation ?)
  	  	  	  	  	  	  	  	  	  // pour depth = 0 contientt les supports
  	  	  	  	  	  	  	  	  	  // des items appartenant à toutes les transactions
  compute_element_support(&support, tt, occs);

  	  	  	  	  	  	  	  	  	  // support[item] non nul =>
  	  	  	  	  	  	  	  	  	  // augmentation + item est "fréquent"
  	  	  	  	  	  	  	  	  	  // dans les transactions occs
  	  	  	  	  	  	  	  	  	  // contenant pattern + augmentation).

  if (depth > 0)
  {
  	  	  	  	  	  	  	  	  	  // ajout de l'augmentation au pattern courant
	  pattern.push_back(pattern_augmentation);
  }

  // simple_pattern_print("pattern               ", pattern, items_names) ;

  int set_support ;

  if (depth > 0)
  {
  	  	  	  	  	  	  	  	  	  // initialise le support du pattern
  	  	  	  	  	  	  	  	  	  // parent + augmentation
  	  	  	  	  	  	  	  	  	  // avec le support de l'augmentation
	  set_support = support[pattern_augmentation];

										  // remet les supports des items appartenant
										  // à pattern + augmentation à 0
	  for(set_t::const_iterator it = pattern.begin(); it != pattern.end(); ++it)
	  {
		if(*it <= max_element)
		  support[*it] = 0;
	  }
  }
  else
  {
	  set_support = occs.size() ;
  }

  // cerr << "occs.size() = " << occs.size() << endl ;
  // cerr << "set_support = " << set_support << endl ;

  	  	  	  	  	  	  	  	  	  // support[item] non null =>
  	  	  	  	  	  	  	  	  	  // (item n'est pas dans pattern)
  	  	  	  	  	  	  	  	  	  // pattern_augmentation + item est present
  	  	  	  	  	  	  	  	  	  // dans les transactions de occs .

						  /***************************************/
						  /* Compute pattern closure             */
  	  	  	  	  	  	  /* patterns dont le support est égal   */
  	  	  	  	  	  	  /* à support[pattern_augmentation]     */
						  /***************************************/

  closure_data_t c_data = {tt, occs, support, set_support};
  set_t closed_pattern = clo(pattern, c_data);
  sort(closed_pattern.begin(), closed_pattern.end());

						  /******************************************************************/
						  /* Check if parent pattern is the first parent of closed_pattern. */
						  /******************************************************************/
  if(!parent_pattern_is_first_parent(closed_pattern, exclusion_list))
  {
    /* If it isn't release the memory and abort the current branch exploration. */
	// cout << "parent pattern is not the first parent of the closed pattern" << endl ;
    return 0;
  }

  nb_closed_patterns++ ;

	// compute the local closures associated with closed_pattern
   int nb_local_closed_patterns2 = 0;


   nb_local_closed_patterns2  = local_closures(depth, tt, occs, support,
		                                       max_element,
		                                       pattern, closed_pattern,
			                                   set_support,
											   threshold_connected_component,
			                                   initial_objects_names,
											   items_names) ;

   if (nb_local_closed_patterns2 == 0)
   {
	   nb_rejected_closed_patterns++ ;
	   return 0 ;
   }

	  	  //TODO could be improved,
for(set_t::const_iterator it = closed_pattern.begin(); it != closed_pattern.end(); ++it)
{
	if(*it <= tt.max_element)
	support[*it] = 0;    //By doing this we remove from db the elements
						// that already belong to the set.
}

 // displayVector("support", support) ;

						  /*************************************************/
						  /* Computes current closed pattern augmentations */
						  /*************************************************/
  set_t augmentations;

  /* to store the value returned by membership foreach element. (Will
     be passed to the rec. call.) */
  set_t augmentations_membership_retval(max_element+1); 

  // cout << "max element = " << max_element << endl ;
  // cout << endl ;
  // displayVector("support", support) ;


  // cerr << "set_support = " << set_support << endl ;

  for(int item = 0; item <= max_element; item++)
  {

    if(support[item] != 0)
    {

      if(!set_member(exclusion_list, item))
      {
    	  augmentations_membership_retval[item] =
    	      	  membership_oracle_on_O2(tt, occs, ot[item]) ;

    	  // cout << "augmentation_membership = " <<   augmentations_membership_retval[item]  << endl ;

    	  if(augmentations_membership_retval[item] > 0)
    	  {
    		  /* Closed_pattern union item is a pattern
    		   * hence item is an augmentation. */
    		  augmentations.push_back(item);
    	  }
      } // fin if (!set member
    } // fin if (support[item]
  } // fin for chaque item
  
  if(augmentations.size() > 0)
  {
    
    if(depth == 0)
      trace_timestamp_print("DBR", EVENT_START ); 
			  
    set_t new_exclusion_list;
    set_t new_el_tail;
    TransactionTable *new_tt;
    TransactionTable *new_ot;

				  /**************************/
				  /* Reuses parent dataset. */
				  /**************************/
      new_tt = &tt;
      new_ot = &ot;

      	  	  	  /* Re-compute the new tids (this can be improved) */
      new_exclusion_list = parent_el; 
      new_el_tail = el_tail; 


    support_sort_cmp_t support_sort = {*new_ot};      
    std::sort(augmentations.begin(), augmentations.end(), support_sort); 

    {
							  /*****************************/
							  /* Recursive call to expand. */
							  /*****************************/
      set_t::const_iterator c_it_end = augmentations.end(); 
      for(set_t::const_iterator c_it = augmentations.begin(); c_it != c_it_end; ++c_it)
      {

    	  Transaction new_occs;

    	  // displayVectorVector("new_ot", *new_ot) ;

		  Transaction new_bis = (*new_ot)[*c_it] ;
		  // displayVector("occs", occs) ;
		  // displayVector("new_bis", new_bis) ;

		  set_intersect(&new_occs, occs, (*new_ot)[*c_it]);


    	  nb_local_closed_patterns2 += expand(*new_tt, *new_ot, new_occs,
                  	  	  	  	              closed_pattern, *c_it, depth+1,
					                          new_exclusion_list, new_el_tail,
                                              // augmentations_membership_retval[*c_it],
				                              threshold_connected_component,
					                          initial_objects_names, items_names,
											  nb_closed_patterns, nb_rejected_closed_patterns);

    	  	  	  	  	  /* insert the current augmentation into the exclusion list for the next calls.*/
    	  new_el_tail.push_back(*c_it);
      }
      // fin for

    }  // fin else
   } // fin if augmentation.size() > 0 ?

  return nb_local_closed_patterns2 ;
}



void paraminer_usage(char *bin_name, bool exit)
{
  cerr<<"ParaMiner general usage (valid for every instance of paraminer):"<<endl;
  cerr<<bin_name<<" [-t <numthreads> = 1] [-c <tuplecutoff> = 2] "<<endl
      <<"[-l (Display tidlist (default is off).)] "<<endl
      <<"[-y <level 0 reduction factor threshold> = 10 (See code doc for details.)] "<<endl
      <<"[-x <level 1 and above reduction factor threshold> = 2 (See code doc for details.)] "<<endl

      <<"<problem specific command line arg 1> .. <problem specific command line arg n>"<<endl<<endl;
  
  if(exit) std::exit(EXIT_FAILURE); 
}

int parse_paraminer_arguments(int *argc, char **argv)
{
  char opt_char=0;
  opterr = 0; 
  // int local_options = 0;
  while ((opt_char = getopt(*argc, argv, "c:x:y:t:l")) != -1)
    {
      switch(opt_char)
	{
	case 'c':
	  if(optarg==NULL || !isdigit(*optarg)) paraminer_usage(argv[0], true);
	  depth_tuple_cutoff = atoi(optarg);
	  *optarg = '\0';
	  cout<<"depth cutoff set to "<<depth_tuple_cutoff<<"."<<endl;
	  break ;
	case 'y':
	  if(optarg==NULL || !isdigit(*optarg)) paraminer_usage(argv[0], true);
	  max_reduction_threshold = atof(optarg);
	  cout<<"max reduction factor threshold set to "<<max_reduction_threshold<<"."<<endl;
	  *optarg = '\0';
	  break; 
	case 'x':
	  if(optarg==NULL || !isdigit(*optarg)) paraminer_usage(argv[0], true);
	  min_reduction_threshold = atof(optarg);
	  cout<<"min reduction factor threshold set to "<<min_reduction_threshold<<"."<<endl;
	  *optarg = '\0';
	  break; 
	case 't':
	  if(optarg==NULL || !isdigit(*optarg)) paraminer_usage(argv[0], true);
	  num_threads = atoi(optarg);
	  *optarg = '\0'; 
	  break;
	case 'l':
	  show_tids = true ;
	  break ;
	default:
	  break;	  
	}
    }

  /* remove options from argv .. */
  for(int i = 0; i < *argc; i++){
    if(!strncmp(argv[i], "-t",2)){
      *(argv[i]) = '\0'; 
    }
    else if (!strncmp(argv[i], "-c",2)){
      *(argv[i]) = '\0'; 
    }
    else if (!strncmp(argv[i], "-x",2)){
      *(argv[i]) = '\0'; 
    }
    else if (!strncmp(argv[i], "-y",2)){
      *(argv[i]) = '\0'; 
    }
    else if (!strncmp(argv[i], "-l",2)){
      *(argv[i]) = '\0' ;
    }
  }

  int idx = 0; 
  for(int i = 0; i < *argc; i++){
    if(*(argv[i]) != '\0'){
      argv[idx++] = argv[i]; 
    }
  }

  *argc = idx; 

  // int tmp = optind; 								// pas utilise
  optind = 0; 
  opterr = 1;   
  cerr<<"running with "<<num_threads<<" thread(s)."<<endl;
  return optind;
}

					// appelé par paraminer_local
int paraminerLC(
				set_t initial_pattern, 				// pas utilisé
				int threshold_connected_component,
				const vector<string>& initial_objects_names,
				const vector<string>& items_names,
				int &nb_closed_patterns,
				int &nb_rejected_closed_patterns
			   )
{

	cerr << endl << "BEGIN PARAMINER LOCAL CLOSURES.." << endl ;

	cout << "==================================================================" << endl ;

	int nb_local_closed_patterns = 0;

	// m_tuplespace_init(&ts, sizeof(tuple_t), 0, TUPLESPACE_OPTIONAUTOCLOSE);
	// m_thread_register();

	set_t empty_set;
	set_t exclusion_list;
	set_t exclusion_list_tail;

	Occurence all_tids(tt.size());

	for(unsigned int i = 0; i < tt.size(); i++)
		all_tids[i]=i;

	SupportTable support(tt.max_element+1, 0);

												// calcule le support
												// de chaque item apparaissant
												// dans les transactions
												// support[item]
												// est le nombre de transactions (sur T)
												// contenant item
	compute_element_support(&support, tt, all_tids);
	// displayVector("support", support) ;


Transaction ot_initial ;

for (unsigned int i = 0 ; i < tt.size() ; i++)
{
	ot_initial.push_back(i) ;
}

nb_local_closed_patterns = expand(
									tt,
									ot,
									ot_initial, // contient les identifiant de toutes les transactions
									empty_set,
									-1,  // pattern_augmentation (rien au depart)
									0,   // depth
									exclusion_list,
									exclusion_list_tail,
									threshold_connected_component,
									initial_objects_names,
									items_names,
									nb_closed_patterns,
									nb_rejected_closed_patterns
								) ;



	cerr << endl <<  "END PARAMINER LOCAL CLOSURES." << endl << endl ;

	return nb_local_closed_patterns;
}

#ifndef TEST
#endif //TEST
