/**
 * @file   paraminer_local_cgraphs.cpp
 * @author Benjamin Negrevergne <benjamin@neb.dyn.cs.kuleuven.be>
 * @date   Tue Oct 19 18:44:38 2010
 * 
 * @brief ParaMiner instance for mining closed frequent itemsets (fim)
 * 
 *
 * PARAMINER LOCAL CLOSURES
 * (by Henry SOLDANO (LIPN University Paris 13)
 *
 * program modified by Dominique Bouthinon (LIPN, University Paris 13)
 *
 */
#include <cstdlib>
#include <algorithm>
#include <vector>
#include <sstream>
#include <fstream>
#include <iostream>
#include <sys/time.h>
#include <time.h>
#include <ctime>
#include <stdio.h>
#include <unistd.h>
#include <set>

#include "paraminer_local.hpp"
#include "pattern.hpp"
#include "database.hpp"
#include "utils.hpp"
#include "paraminer_LC2.hpp"
// #include "matrix.hpp"
#include "read_write_data.hpp"
#include "display_data.hpp"
#include "utils.hpp"


using std::cout; 
using std::cerr; 
using std::endl; 


#define PERMUTE_ITEMS


int ELEMENT_RANGE_END; 

extern int threshold;

set_t permutations;

using namespace std;
using namespace boost ;


/**
 *
 * (appelé depuis paraminer (et non pas paraminer_local_itemsets))
 */

void element_print(const element_t element)
{
  cout<< permutations[element] ;
}


void original_element_print(const element_t element,
				   	        const vector<string>& items_names)
{

  if (element >= permutations.size())
  {
	  cout << "problem" << endl ;
  }

  if (permutations[element] >= items_names.size())
  {
	  cout << "problem 2" << endl ;
  }

  // cout << "element = " << element << endl ;
  // cout << "permutation[element] = " << permutations[element] << endl ;
  cout<< items_names[permutations[element]] ;
}


typedef struct
{
  bool operator()(const pair<int,int> &a, const pair<int,int> &b)
  {
    return a.second < b.second; 
  }
}
pair_gt_second_t;


// calcul de la frequence des items
// sur les objets initiaux

vector<pair<int,int> >
compute_frequences_initial_objects(
                                   const vector<vector<int> >& item_occurrences,
	     	 	 	 	 	 	   const vector<vector<int> >& initial_objects
								  )
{
	vector<pair<int,int> > frequences ;

									// pour chaque item
	for (unsigned int item = 0 ; item < item_occurrences.size() ; item++)
	{
		set<int> objets ;			// objets initiaux contenant l'item

									// item_occurrences.at(item) est la
									// liste des numeros transactions
									// contenant l'item

		// cout << "item = " << item << endl ;
		// displayVector("trans item", item_occurrences.at(item)) ;

									// pour chaque transaction
									// contenant l'item
		for (unsigned int i = 0 ; i < item_occurrences.at(item).size() ; i++)
		{
			int transaction = item_occurrences.at(item).at(i) ;

			// cout << "item = " << item << endl ;
			// cout << "transaction = " << transaction << endl ;
			// displayVector("objets", initial_objects.at(transaction)) ;

									// objets_initiaux.at(transaction) est
									// la liste des objets initiaux associes
									// à l'objet décrit par la transaction
			for (unsigned int j = 0 ;
				 j < initial_objects.at(transaction).size() ; j++)
			{

									// objet initial contenant l'item
				int objet = initial_objects.at(transaction).at(j) ;

				// cout << "objet = " << objet << endl ;

                objets.insert(objet) ;

                // displayVector("objets", objets) ;
			}
			// end for int j
		}
		// end for int i
									// objets.size() est le nombre d'objets
									// initiaux contenant item

		// displayVector("objets", objets) ;
		int frequence_objet = objets.size() ;
		frequences.push_back(pair<int,int>(item, frequence_objet)) ;
	}
	// end for in item

	return frequences ;
}
// end compute_frequences_initial_objects



/*   calcule la frequence des items
 * 	 ne retient que les items dont la frequence est superieure ou
 * 	 egale a minsup
 * 	 Attention ! : la frequence doit être calculee sur les objets initiaux O
 * 	 et non pas sur transactions.
 * 	 ordonne les items par frequence croissant
 * 	 permutation[item] est un indice attribué à l'item (conservé)
 */
int compute_permutation_by_frequency(
										vector<vector<int> >& transacts,
										set_t *permutations,
										int minsup,
										vector<vector<int>>& initial_objects
									 )
{

	vector<vector<int> > item_occurrences ;
								// item_occurrences[item] contient les numéros
								// des transactions contenant item

	element_t max_element = -1;

								// Pour chaque transaction
	for (unsigned int i = 0 ; i < transacts.size() ; i++)
	{
		vector<int>& transact = transacts.at(i) ;

		// displayVector("transact", transact) ;

								// pour chaque item de la transaction
		for (unsigned int j = 0 ; j < transact.size(); j++)
		{
			int& item = transact.at(j) ;

			// cout << "item = " << item << endl ;

			if(item > max_element)		// 1ere rencontre de l'item
			{
				  max_element=item;
				  item_occurrences.resize(max_element+1);
			}
			  	  	  	  	  	  	  	  //  item apparait dans transaction i
			item_occurrences[item].push_back(i);
		 }
	}

	// displayVectorVector("item_occurrences", item_occurrences) ;

	vector<pair<int,int> > freq_table = compute_frequences_initial_objects
			                               (item_occurrences, initial_objects) ;

	 pair_gt_second_t gt;
	 sort(freq_table.begin(), freq_table.end(),gt);

						  // freq_table est trie en ordre croissant selon
						  // la frequence des items

	permutations->resize(max_element+1);

						 // on attribue a chaque item un index dans
						 // permutation
						 // en ne conservant que les items suffisamment
						 // representés
	int item_index = 0;

	for(unsigned int i = 0; i < freq_table.size(); i++)
	{
		// cout << "item  = " << freq_table.at(i).first
		//	 << " freq  = " << freq_table.at(i).second << endl ;

		if(freq_table[i].second >= minsup)
		  (*permutations)[freq_table[i].first] = item_index++;
		else
		  (*permutations)[freq_table[i].first] = -1;
	}


	#ifdef PERMUTE_ITEMS

	#else
				  /* reset permutations (dirty) */
	for(int i = 0; i < permutations->size(); i++)
	{
		(*permutations)[i] = i;
	}
	#endif

	return item_index;
}




		// teste si l'extension du pattern courant (base + extension)
		// est suffisante (superieur ou égale à threshold)
int membership_oracle(
					  const set_t &base_set, 				// pas utilisé
					  const element_t extension,
		      	  	  const membership_data_t &data
					  )
{
  return data.support[extension] >= threshold ? data.support[extension] : 0; 
}


	// teste si u(ext( closed_pattern + extension)) >= threshold
int membership_oracle_on_O2(
							const TransactionTable &tt,
							const Occurence &closed_pattern_occurences,
							const Occurence &extension_occurences
			  	  	       )
{
	Occurence occ ;

	// displayVector("closed", closed_pattern_occurences) ;
	// displayVector("extension", extension_occurences) ;

	set_intersect(&occ, closed_pattern_occurences, extension_occurences) ;

	// displayVector("occ", occ) ;

	set<int> objets ;		// objets initiaux contenant item

	// displayVector("base_set", base_set) ;
	// displayVector("base_set occurrences ", data.base_set_occurences) ;


	// cout << "extension = " << extension << endl ;
	// displayVector("extension occurrences ", data.extension_occurences) ;

								// recupere les objets initiaux
								// associés à chaque transaction
								// contenant extension (+ base_set)
	for (unsigned int i=0 ; i < occ.size() ; i++)
	{
		int tid = occ.at(i) ;

		// cerr << "tid = " << tid << endl ;

		/*

		for (int j = 0 ; j < tt[tid].initial_objects.size() ; j++)
		{
			cerr << tt[tid].initial_objects.at(j) << endl ;
		}

		*/


								// tid identifie une transaction dans tt
		objets.insert(tt[tid].initial_objects.begin(),
					  tt[tid].initial_objects.end()) ;
	}

	int support = objets.size() ;

	if (support >= threshold)
		return support ;

	return 0;
}



set_t clo(const set_t &set, 					// pattern à clore
		  const closure_data_t &data)
{
  return support_based_closure(set, data.set_support, data.support); 
}


/**
 *  affecte une liste d'objet initiaux à chaque transaction
 */
/*
void transactions_initial_objects(
								  TransactionTable& tt,
								  matrix<element_t> objects
								 )
{
	// cout << tt.size() << endl ;
									// pour chaque Transaction
	for (unsigned int i = 0 ; i < tt.size() ; i++)
	{
		tt.at(i).initial_objects = objects.at(i) ;
	}
}
*/



element_t build_transaction_table(TransactionTable *tt,
							     const vector<vector<int> >& transacts,
								 const set_t& permutations)
{
	element_t max_element = 0;

	int nb_items = 0;
	int nb_trans = 0;

	  	  	  	  	  	  	  	  // pour chaque transaction de transacts
	for (unsigned int i = 0 ; i < transacts.size() ; i++)
	{
		Transaction t;

		// displayVector("transact", transacts.at(i)) ;

		// cout << "transaction : " << i << endl ;

								// pour chaque item de transact
		for (unsigned int j = 0 ; j < transacts.at(i).size() ; j++)
		{
			  // cout << "         indice item = " << j << endl ;

			  int item = transacts.at(i).at(j) ;

			  // cout << "         item = " << item << endl ;

			  element_t p = permutations[item];

			  // cout << "         p = " << p << endl ;

			  // cout << "         max_element = " << max_element << endl ;

			  // cout << "p = " << p << endl ;

			  if(p != -1)
			  {
				  t.push_back(p);
			  }

			  if(p > max_element)
				  max_element=p;

		    nb_items++ ;
		}

		// if(t.size() != 0)
		// {
			t.limit=t.size();
			#ifdef TRACK_TIDS
			  if(show_tids)
				  t.tids.push_back(nb_trans);
			#endif //TRACK_TIDS
		   nb_trans++;
		   t.weight = 1;
		   #ifdef SORT_DATABASE
				  sort(t.begin(), t.end());
		   #else
		   #ifndef NDEBUG
			  assert(is_sorted(t));
	   	   #endif
		   #endif //SORT_DATABASE
		   tt->push_back(t);
		 //}
	  }
	  // fin pour chaque transaction

	  cout  <<  "Data loaded, "<<nb_items<<" items within "
			<< nb_trans<<" transactions."<<endl;

	  tt->max_element = max_element;

	  return max_element;
}


// add the adjacences to the transactions
void add_adjacences_initial_objects(
									TransactionTable *tt,
									vector<vector<int> >& adjacences,
									const vector<vector<int> >& initial_objects
									)
{

	// displayVectorVector("initial_objects", initial_objects) ;

									// pour chaque transation
	for(unsigned int i=0 ; i < tt->size() ; i++)
	{
									// on trie la adjacences pour éviter
									// les noeuds en double dans la suite
									// du programme
		sort(adjacences.at(i).begin(), adjacences.at(i).end()) ;
		tt->at(i).adjacences      = adjacences.at(i) ;
		tt->at(i).initial_objects = initial_objects.at(i) ;

		// displayVector("initial_objects (dans)", tt->at(i).initial_objects) ;
	}
}




void print_arguments(int argc, char **argv)
{
	cout << "arguments : " << endl ;
	for (int i = 0 ; i < argc ; i++)
	{
		cout << i << " : " << argv[i] << endl ;
	}

	cout << "fin arguments " << endl << endl ;
}



void usage(char *bin_name)
{
  paraminer_usage(bin_name, false);
  cerr<<"Problem specific command line arguments:"<<endl;
  cerr<<bin_name<<" [<paraminer_options> (See above.)] <dataset> <minsup>"<<endl;
  exit(EXIT_FAILURE);
}


// ============================== MAIN ========================================

int main(int argc, char **argv)
{

  parse_paraminer_arguments(&argc, argv);

  int threshold_connected_component ; 			// threshold to keep a connected
	  	  	  	  	  	  	  	  	  	  	  	// component
  vector<string> initial_objects_names ;
  vector<string> items_names ;
  vector<vector<int> > initial_objects ;
  vector<vector<int> > transacts ;
  vector<vector<int> > adjacences ;

  int nb_closed_patterns          = 0 ;
  int nb_rejected_closed_patterns = 0 ;

  set_t empty_set;


  threshold = std::atoi(argv[2]);
  threshold_connected_component = std::atoi(argv[3]) ;

  cout << "data file                            = " << argv[1]   << endl ;
  cout << "minimal frequency of closed patterns = " << threshold << endl ;
  cout << "minimal cardinality of communities   = " << threshold_connected_component << endl ;


  cerr << endl << "readind data..." << endl ;
  	  	  	  	  	  	  	  	  	  	  	  // read all data in file argv[1]
  read_data(argv[1], initial_objects_names, items_names, initial_objects,
		    transacts, adjacences) ;

  // displayVector("object_names", object_names) ;
  // displayVector("item_names", item_names) ;
  // displayVectorVector("initial_objects", initial_objects) ;

  // displayVectorVector("transacts", transacts) ;
  // displayVectorVector("adjacences", adjacences) ;


  // deleteIrrelevantAdjacences(adjacences) ;

  cerr << "computing permutations..." << endl ;

  compute_permutation_by_frequency(transacts,
		  	  	  	  	  	  	   &permutations,
								   threshold,
								   initial_objects);

  // displayVector("permutations", permutations) ;

  element_t max = build_transaction_table(&tt, transacts, permutations);

  // print_transaction_table(tt) ;

						  // attention des items non frequents ont été éliminés
						  // il ne faut donc plus les considérer

 add_adjacences_initial_objects(&tt, adjacences, initial_objects) ;

  	  	  	  	  	  	  // inverse indices et valeurs dans permutations
 reverse_permutations(&permutations);

 // displayVector("permutations ", permutations) ;


  // print_transaction_table(tt) ;
  // print_transaction_table_attributes(tt,permutations) ;

  ELEMENT_RANGE_END = max+1;

  // print_transaction_table(tt);

  // Occurence tids; 
  // for(int i = 0; i < tt.size(); i++){
  //   tids.push_back(i); 
  // }
  // quick_sort_tids(tt, &tids, 0, tids.size());
  // set_print(tids);

  cerr << "tranposing..." << endl ;

  transpose(tt, &ot);

  trace_init(num_threads);
  
  time_t debut ;
  time(&debut) ;


  int num_pattern = paraminerLC(empty_set, threshold_connected_component,
		  	  	  	  	  	    initial_objects_names, items_names,
								nb_closed_patterns, nb_rejected_closed_patterns);

  cout << "===================================================================" << endl ;
  cout << nb_closed_patterns          << " closed pattern(s) (#c)"          << endl ;
  cout << nb_rejected_closed_patterns << " rejected closed pattern(s) (they have no satisfying local closed pattern)" << endl ;
  cout << num_pattern                 << " triplets (# (c, e, l))."    << endl ;

  time_t fin ;
  time(&fin) ;

  cout << "begin      = " << ctime(&debut) ;
  cout << "end        = " << ctime(&fin)  ;
  cout << "duration   = " << difftime(fin,debut) << " seconds." << endl ;

  return 0 ;

}
